########################################################################
#
# File Name:            DbmDatabase.py
#
# Documentation:        http://docs.4suite.org/Lib/DbmDatabase.py.html
#
"""
A utility class for anydbm databases.
WWW: http://4suite.org/         e-mail: support@4suite.com

Copyright (c) 1999 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import os, string, anydbm, shutil, sys
from distutils import dir_util
from cPickle import dumps, loads

from Ft.Lib import Set

__version__ = '0.5.0'

class DbmError(Exception):
    def __init__(self, cmd, msg):
        self.cmd = cmd
        self.msg = msg

    def __str__(self):
        return '%s: %s' % (self.cmd, self.msg)

FT_DATABASE = 'ft_database'

if os.name == 'posix':
    DATABASE_DIR = '/var/local/data/ftdatabase'
else:
    DATABASE_DIR = None

DATABASE_DIR = os.environ.get('FT_DATABASE_DIR', DATABASE_DIR)

if DATABASE_DIR is None:    
    raise DbmError('BOOTSTRAP', 'Please define FT_DATABASE_DIR in the environment')

class Table:

    def __init__(self,name,dbFile):
        self.__name = name
        self.__dbFile = dbFile
        self.__db = None
        self.__inTx = 1

        self.__cached = {}
        self.__added = {}
        self.__deleted = []

    def getName(self):
        return self.__name


    def rollback(self):
        self.__verify()
        self.__inTx = 0
        if self.__db:
            self.__db.close()


    def checkpoint(self):
        self.__verify()
        if not self.__db:
            self.__db = anydbm.open(self.__dbFile,CREATE)
        for k,v in self.__cached.items():
            self.__db[k] = dumps(v)
        for k,v in self.__added.items():
            self.__db[k] = dumps(v)
        for k in self.__deleted:
            if self.__db.has_key(k):
                del self.__db[k]

    def commit(self):
        self.checkpoint()
        self.__inTx = 0
        if self.__db:
            self.__db.close()

    def __verify(self):
        if not self.__inTx:
            raise DbmError('TRANSACTION ERROR',"Transaction not in progress")
        if os.path.exists(self.__dbFile) and self.__db is None:
            self.__db = anydbm.open(self.__dbFile, WRITEABLE)
        

    def keys(self):
        self.__verify()
        keys =  (self.__db and self.__db.keys()  or []) + self.__added.keys()
        for k in self.__deleted:
            if k in keys:
                keys.remove(k)
        return keys

    def values(self):
        return map(lambda x,s=self:s[x],self.keys())
    def items(self):
        return map(lambda x,s=self:(x,s[x]),self.keys())

    def has_key(self, key):
        self.__verify()
        return key in self.keys()

    def get(self, key, default=None):
        self.__verify()
        if self.has_key(key):
            return self[key]
        return default

    def __len__(self):
        return len(self.keys())

    def __getitem__(self, key):
        self.__verify()
        if key in self.__deleted:
            raise KeyError, key
        if self.__added.has_key(key):
            return self.__added[key]
        if self.__cached.has_key(key):
            return self.__cached[key]
        if self.__db and self.__db.has_key(key):
            o = self.__db[key]
            self.__cached[key] = loads(o)
            return self.__cached[key]
        raise KeyError(key)

    def __setitem__(self, key, value):
        self.__verify()
        self.__added[key] = value
        if key in self.__deleted: self.__deleted.remove(key)

    def __delitem__(self, key):
        self.__verify()
        if not self.has_key(key):
            raise KeyError, key
        self.__deleted.append(key)
        if self.__added.has_key(key): del self.__added[key]

    def __del__(self):
        try:
            if self.__inTx:
                self.rollback()
        except:
            pass


CREATE = 'c'
NEW = 'n'
WRITEABLE = 'w'
READONLY = 'r'

def SysInit():
    dir_util.mkpath(DATABASE_DIR)
    version_file = os.path.join(DATABASE_DIR, 'VERSION')
    if os.path.exists(version_file):
        raise "Already a database at %s" % DATABASE_DIR
    dir_util.mkpath(DATABASE_DIR)
    f = open(version_file,'w')
    f.write(__version__)
    f.close()

    #Create an initial DB
    dir_util.mkpath(os.path.join(DATABASE_DIR,'template1'))

try:
    SysInit()
except:
    pass

def CreateDatabase(dbName):
    CheckVersion()
    dbpath = os.path.join(DATABASE_DIR, dbName)
    if os.path.exists(dbpath):
        raise DbmError('CREATE DATABASE', 'Database "%s" already exists' % dbName)

    dir_util._path_created = {}
    dir_util.mkpath(dbpath)
    return Database(dbpath)

def DropDatabase(dbName):
    CheckVersion()
    dbpath = os.path.join(DATABASE_DIR, dbName)
    if not os.path.exists(dbpath):
        raise DbmError('DROP DATABASE', 'Database "%s" does not exists' % dbName)

    #Delete all of the files
    for f in os.listdir(dbpath):
        os.unlink(os.path.join(dbpath,f))
    os.rmdir(dbpath)
    

def GetAllDatabaseNames():
    #Get all from the file system
    dbs = []
    CheckVersion()
    for n in os.listdir(DATABASE_DIR):
        if os.path.isdir(os.path.join(DATABASE_DIR,n)):
            dbs.append(n)
    return Set.Unique(dbs)



def CheckVersion():
    version_file = os.path.join(DATABASE_DIR, 'VERSION')
    if not os.path.exists(version_file):
        raise "No database at %s" % DATABASE_DIR
    f = open(version_file,'r')
    if f.read() != __version__:
        raise "Wrong Database version"

class Database:
    """Each instance of a database represents a single transactionm
    """
    def __init__(self, name):
        """
        initialize the database object.
        """
        self._name = name
        CheckVersion()
        self._dbpath = os.path.join(DATABASE_DIR, name)

        if not os.path.exists(self._dbpath):
            raise DbmError('CONNECT DATABASE', 'Database "%s" does not exist' % name)

        self.__tables = {}
        self._inTx = 1

    def __clear(self):
        self.__tables = {}

    def getName(self):
        return self._name

    def createTable(self, table):
        if not self._inTx:
            raise DbmError('TRANSACTION ERROR', 'Transaction Not in Progress')
        table_file = os.path.join(self._dbpath, table+'.db')
        if os.path.exists(table_file):
            raise DbmError('CREATE TABLE', 'Table Already exists' % table)
            
        t = Table(table,table_file)
        self.__tables[table] = t
        return t

    def getAllTableNames(self):
        tbs = map(lambda x:x.getName(),self.__tables.values())
        #Get all from the file system
        tbs = tbs + map(lambda x:os.path.splitext(x)[0],os.listdir(self._dbpath))
        return Set.Unique(tbs)

    keys = getAllTableNames

    def getTable(self, table):
        if self.__tables.has_key(table):
            return self.__tables[table]
        table_file = os.path.join(self._dbpath, table+'.db')
        if not os.path.exists(table_file):
            raise DbmError('GET TABLE', 'Table %s does not exists' % table)
            
        t = Table(table,table_file)
        self.__tables[table] = t
        return t

    __getitem__ = getTable
    
    def dropTable(self, table):
        table_file = os.path.join(self._dbpath, table+'.db')
        if os.path.exists(table_file):
            os.remove(table_file)
            return 1
        return 0

    __delitem__ = dropTable


    def rollback(self):
        #Tell all of our tables to roll back
        for t in self.__tables.values():
            t.rollback()
        self.__clear()
        self._inTx = 0

    def checkpoint(self):
        #Tell all of our tables to roll back
        for t in self.__tables.values():
            t.checkpoint()

    def commit(self):
        for t in self.__tables.values():
            t.commit()
        self.__clear()
        self._inTx = 0

    def __del__(self):
        try:
            if self._inTx:
                self.rollback()
        except:
            pass


if __name__ == '__main__':
    import sys
    if len(sys.argv) == 4:
        d = Database(sys.argv[1])
        t = d[sys.argv[2]]
        print t[sys.argv[3]]
    elif len(sys.argv) == 3:
        d = Database(sys.argv[1])
        t = d[sys.argv[2]]
        print t.keys()
    elif len(sys.argv) == 2:
        d = Database(sys.argv[1])
        print d.getAllTableNames()
    else:
        print GetAllDatabaseNames()
