########################################################################
#
# File Name:            Domlette.py
#
# Documentation:        http://docs.4suite.com/4XSLT/Domlette.py.html
#
"""
A tiny, read-only DOM optimized for XPath.  Domlette is also automatically
normalized
WWW: http://4suite.com/4XSLT        e-mail: support@4suite.com

Copyright (c) 2000 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""

import sys, string, os, urllib
import UserDict, cStringIO, cPickle
from xml.dom.ext import SplitQName
from Ft.Lib import ReaderBase, Uri
from xml.dom import XML_NAMESPACE, XMLNS_NAMESPACE

from Ft.Lib import FtException, Error

from types import StringType
stringTypes = [StringType]
try:
    from types import UnicodeType
    stringTypes.append(UnicodeType)
except ImportError, e:
    pass

import Ft.Lib
from xml.parsers import expat

DOM_IMP = Ft.Lib.PDOMLETTE


class Node:
    # Node types
    ELEMENT_NODE                = 1
    ATTRIBUTE_NODE              = 2
    TEXT_NODE                   = 3
    CDATA_SECTION_NODE          = 4
    ENTITY_REFERENCE_NODE       = 5
    ENTITY_NODE                 = 6
    PROCESSING_INSTRUCTION_NODE = 7
    COMMENT_NODE                = 8
    DOCUMENT_NODE               = 9
    DOCUMENT_TYPE_NODE          = 10
    DOCUMENT_FRAGMENT_NODE      = 11
    NOTATION_NODE               = 12

    def __init__(self, ownerDocument, nodeType, namespaceURI='', localName='', prefix=''):
        self.childNodes = []
        self.nodeName = ''
        self.nodeValue = ''
        self.parentNode = None
        self.firstChild = None
        self.lastChild = None
        self.previousSibling = None
        self.nextSibling = None
        self.attributes = None
        self.ownerDocument = ownerDocument
        self.namespaceURI = namespaceURI
        self.prefix = prefix
        self.localName = localName
        self.readOnly = 0
        self.nodeType = nodeType

        return

    def hasChildNodes(self):
        return childNodes and 1 or 0

    def normalize(self):
        pass

    def appendChild(self, newChild):
        if (newChild.nodeType == Node.DOCUMENT_FRAGMENT_NODE and
            newChild.firstChild):
            # Only for non-empty document fragments
            for node in newChild.childNodes:
                node.parentNode = self
            if self.lastChild:
                # Graft the docfrag's childNodes to the end of ours
                self.lastChild.nextSibling = newChild.firstChild
                self.lastChild = newChild.lastChild
                newChild.firstChild.previousSibling = self.lastChild
                self.childNodes.extend(newChild.childNodes)
            else:
                # We are empty, just take the whole thing
                self.childNodes = newChild.childNodes
                self.firstChild = newChild.firstChild
                self.lastChild = newChild.lastChild

            # Empty the document fragment
            newChild.childNodes = []
            newChild.firstChild = None
            newChild.lastChild = None
        else:
            if newChild.parentNode:
                newChild.parentNode.removeChild(newChild)
            newChild.parentNode = self
            newChild.ownerDocument = self.ownerDocument
            if not self.firstChild:
                self.firstChild = newChild
            else:
                self.lastChild.nextSibling = newChild
                newChild.previousSibling = self.lastChild
            self.lastChild = newChild
            self.childNodes.append(newChild)
        return newChild

    def removeChild(self, childNode):
        self.childNodes.remove(childNode)

        if childNode.nextSibling:
            childNode.nextSibling.previousSibling = childNode.previousSibling
        else:
            self.lastChild = childNode.previousSibling

        if childNode.previousSibling:
            childNode.previousSibling.nextSibling = childNode.nextSibling
        else:
            self.firstChild = childNode.nextSibling

        childNode.parentNode = None
        childNode.previousSibling = None
        childNode.nextSibling = None
        return childNode

    def replaceChild(self, newChild, oldChild):
        sibling = oldChild.nextSibling
        self.removeChild(oldChild)
        if sibling:
            self.insertBefore(newChild, sibling)
        else:
            self.appendChild(newChild)
        return oldChild

    def insertBefore(self, newChild, refChild):
        if refChild is None:
            return self.appendChild(newChild)
        else:
            index = self.childNodes.index(refChild)
            if (newChild.nodeType == Node.DOCUMENT_FRAGMENT_NODE and
                newChild.firstChild):
                # Only for non-empty document fragments
                for node in newChild.childNodes:
                    node.parentNode = self
                if index:
                    # It's in the middle
                    refChild.previousSibling.nextSibling = newChild.firstChild
                    newChild.firstChild.previousSibling = refChild.previousSibling
                    refChild.previousSibling = newChild.lastChild
                    newChild.lastChild.nextSibling = refChild
                    children = self.childNodes[:index]
                    children.extend(newChild.childNodes)
                    children.extend(self.childNodes[index:])
                    self.childNodes = children
                else:
                    # At the beginning
                    newChild.childNodes.extend(self.childNodes)
                    newChild.lastChild.nextSibling = self.firstChild
                    self.firstChild.previousSibling = newChild.lastChild
                    self.firstChild = newChild.firstChild
                    self.childNodes = newChild.childNodes
                # Empty out the document fragment
                newChild.childNodes = []
                newChild.firstChild = None
                newChild.lastChild = None
            else:
                self.childNodes.insert(index, newChild)
                if newChild.parentNode:
                    newChild.parentNode.removeChild(newChild)
                newChild.parentNode = self
                if refChild.previousSibling:
                    refChild.previousSibling.nextSibling = newChild
                else:
                    self.firstChild = newChild
                
                newChild.previousSibling = refChild.previousSibling
                refChild.previousSibling = newChild
                newChild.nextSibling = refChild
            return newChild

    def isSameNode(other):
        return self == other


class Document(Node):
    def __init__(self):
        Node.__init__(self, self, Node.DOCUMENT_NODE)
        self.documentElement = None
        self.doctype = None

    def createElementNS(self, namespaceURI, qualifiedName):
        prefix, localName = SplitQName(qualifiedName)
        return Element(self, namespaceURI, localName, prefix)

    def createAttributeNS(self, namespaceURI, qualifiedName):
        prefix, localName = SplitQName(qualifiedName)
        if localName == 'xmlns':
            return Attribute(self, XMLNS_NAMESPACE, prefix, 'xmlns')
        else:
            return Attribute(self, namespaceURI, localName, prefix)

    def createTextNode(self,data):
        t = Text(self)
        t.data = data
        return t

    def createDocumentFragment(self):
        df = DocumentFragment(self)
        return df

    def createNodeIterator(self, root, whatToShow, filter, entityReferenceExpansion):
        from xml.dom import NodeIterator
        nodi = NodeIterator.NodeIterator(root, whatToShow, filter, entityReferenceExpansion)
        return nodi

    def appendChild(self, newChild):
        if not self.documentElement and newChild.nodeType == Node.ELEMENT_NODE:
            self.documentElement = newChild
        return Node.appendChild(self, newChild)

    def isHtml(self):
        return 0

    def isXml(self):
        return 1

    #Range support
    def createRange(self):
        from xml.dom import Range
        return Range.Range(self)


    def __repr__(self):
        return '<Domlette Document Node at %x>' % id(self)

    def __getinitargs__(self):
        raise Exception("You should not pickle the document")

class Element(Node):
    def __init__(self, ownerDocument, namespaceURI='', localName='', prefix=''):
        Node.__init__(self, ownerDocument, Node.ELEMENT_NODE, namespaceURI,
                      localName, prefix)
        self.attributes = NamedNodeMap({})
        self.nodeName = prefix and prefix + ':' + localName or localName
        self.tagName = self.nodeName

    def getAttributeNodeNS(self, namespaceURI, localName):
        return self.attributes.get((namespaceURI, localName), None)

    def getAttributeNS(self, namespaceURI, localName):
        attr = self.getAttributeNodeNS(namespaceURI, localName)
        return attr and attr.value or ''

    def setAttributeNS(self, namespaceURI, qualifiedName, value):
        attr = self.ownerDocument.createAttributeNS(namespaceURI,
                                                    qualifiedName)
        attr.value = value
        attr.ownerElement = self
        prefix, localName = SplitQName(qualifiedName)
        if localName == 'xmlns':
            self.attributes[(XMLNS_NAMESPACE, prefix)] = attr
        else:
            self.attributes[(namespaceURI, localName)] = attr
        self.attributes.length = self.attributes.length + 1
        return

    def setAttributeNodeNS(self,attrNode):
        return self.setAttributeNS(attrNode.namespaceURI,attrNode.name,attrNode.value)

    def removeAttributeNode(self,attr):
        for n,a in self.attributes.items():
            if a == attr:
                del self.attributes[n]
                return

    def hasAttributeNS(self, namespaceURI, localName):
        attr = self.attributes.get((namespaceURI, localName), None)
        return attr and 1 or 0

    def cloneNode(self,deep):
        newElement = self.ownerDocument.createElementNS(self.namespaceURI,self.nodeName)
        for attr in self.attributes.values():
            newElement.setAttributeNS(attr.namespaceURI,attr.name,attr.value)
        if deep:
            for c in self.childNodes:
                nc = c.cloneNode(deep)
                newElement.appendChild(nc)
            
        return newElement
    
    def __repr__(self):
        return "<Domlette Element Node at %x: name='%s' with %d attributes and %d children>" % (
            id(self),
            self.nodeName,
            len(self.attributes),
            len(self.childNodes)
            )

    def __getinitargs__(self):
        return (None, self.namespaceURI, self.localName, self.prefix)

    def __getstate__(self):
        attrs = []
        for attr in self.attributes.values():
            attrs.append((attr.namespaceURI, attr.localName, attr.prefix,
                          attr.value))
        return attrs

    def __setstate__(self, attrs):
        for data in attrs:
            attr = Attribute(self.ownerDocument, data[0], data[1], data[2])
            attr.value = data[3]
            if data[1] == 'xmlns':
                self.attributes[(data[0], data[2])] = attr
            else:
                self.attributes[(data[0], data[1])] = attr
        return


class Attribute(Node):
    def __init__(self, ownerDocument, namespaceURI='', localName='', prefix=''):
        Node.__init__(self, ownerDocument, Node.ATTRIBUTE_NODE, namespaceURI, localName, prefix)
        self.ownerElement = None
        if not prefix:
            self.name = localName
        elif not localName:
            self.name = prefix
        else:
            self.name = prefix + ':' + localName
        self.nodeName = self.name
        self.value = ''
        return

    def __repr__(self):
         return "<Domlette Attribute Node at %x: name='%s', value='%s'>" % (
             id(self),
             self.name,
             self.value
             )

    def __getinitargs__(self):
        return (None, self.namespaceURI, self.localName, self.prefix)

    def __getstate__(self):
        return {'value':self.value}


class DocumentFragment(Node):
    def __init__(self, ownerDocument):
        Node.__init__(self, ownerDocument, Node.DOCUMENT_FRAGMENT_NODE)
        self.nodeName = '#document-fragment'
        return

    def __repr__(self):
        return '<Domlette DocumentFragment Node at %x with %d children>' % (
                id(self),
                len(self.childNodes)
                )

    def __getinitargs__(self):
        raise Exception("You should not pickle a document fragment")


class Text(Node):
    def __init__(self, ownerDocument):
        Node.__init__(self, ownerDocument, Node.TEXT_NODE)
        self.nodeName = '#text'
        self.data = ''

    def deleteData(self,offset,count):
        self.data = self.data[:int(offset)] + self.data[int(offset+count):]

    def substringData(self, offset, count):
        return self.data[int(offset):int(offset+count)]

    def cloneNode(self,deep):
        return self.ownerDocument.createTextNode(self.data)
        

    def __repr__(self):
        # Trim to a managable size
        if len(self.data) > 20:
            data = self.data[:20] + '...'
        else:
            data = self.data

        # Escape unprintable chars
        import string
        for ws in ['\t','\n','\r']:
            data = string.replace(data, ws, '\\0x%x' % ord(ws))

        st = "<Domlette Text Node at %x: data='%s'>" % (
                id(self),
                data
                )
        return st

    def __getinitargs__(self):
        return (None,)

    def __getstate__(self):
        return {'data':self.data}


class ProcessingInstruction(Node):
    def __init__(self, ownerDocument):
        Node.__init__(self, ownerDocument, Node.PROCESSING_INSTRUCTION_NODE)
        self.nodeName = '#processing-instruction'
        self.target = ''
        self.data = ''

    def __repr__(self):
        return "<Domlette Processing Instruction Node at %x: target='%s%s', data='%s%s'>" % (
            id(self),
            self.target[:20],
            len(self.target) > 20 and "..." or "",
            self.data[:20],
            len(self.data) > 20 and "..." or ""
            )

    def __getinitargs__(self):
        return (None,)

    def __getstate__(self):
        return {'target':self.target, 'data':self.data}


class Comment(Node):
    def __init__(self, ownerDocument):
        Node.__init__(self, ownerDocument, Node.COMMENT_NODE)
        self.nodeName = '#comment'
        self.data = ''

    def __repr__(self):
        return "<Domlette Comment Node at %x: data='%s%s'>" % (
            id(self),
            self.data[:20],
            len(self.data) > 20 and "..." or ""
            )

    def __getinitargs__(self):
        return (None,)

    def __getstate__(self):
        return {'data':self.data}


class NamedNodeMap(UserDict.UserDict):
    length = 0
    def __init__(self, ownerDocument=None):
        UserDict.UserDict.__init__(self)
        self.ownerDocument = ownerDocument

    def item(self, index):
        try:
            return self[self.keys()[int(index)]]
        except IndexError:
            return None

    def __getitem__(self, index):
        if type(index) == type(0):
            return self[self.keys()[index]]
        else:
            return UserDict.UserDict.__getitem__(self, index)

    def __repr__(self):
        st = "<Domlette NamedNodeMap at %x: {"% id(self)
        st = reduce(lambda s, (k,v): s+repr(k)+':'+repr(v)+', ', self.items(), st)
        if len(self):
            st = st[:-2]
        return st + '}>'


def ReleaseNode(node):
    for child in node.childNodes:
        ReleaseNode(child)
    node.childNodes = []
    if node.nodeType == Node.ELEMENT_NODE:
        for attr in node.attributes.values():
            ReleaseNode(attr)
        node.attributes = {}
    node.parentNode = None
    node.firstChild = None
    node.lastChild = None
    node.previousSibling = None
    node.nextSibling = None
    node.ownerDocument = None
    if node.nodeType == Node.ATTRIBUTE_NODE:
        node.ownerElement = None
    return


class Utf8OnlyHandler:
    def __init__(self, origReader):
        try:
            import codecs
        except:
            raise Exception("Don't use force8Bit on Python versions that don't support unicode")
        self.encoder = codecs.lookup('utf-8')[0]
        self.origReader = origReader
        return

    def enc(self, st):
        return self.encoder(st)[0]
    
    def processingInstruction(self, target, data):
        self.origReader.processingInstruction(self.enc(target), self.enc(data))
        return

    def comment(self, data):
        self.origReader.comment(self.enc(data))
        return

    def startElement(self, name, attribs):
        new_attribs = {}
        for k in attribs.keys():
            new_attribs[self.enc(k)] = self.enc(attribs[k])
        self.origReader.startElement(self.enc(name), new_attribs)
        return

    def endElement(self, name):
        self.origReader.endElement(self.enc(name))
        return

    def characters(self, data):
        self.origReader.characters(self.enc(data))
        return

    def entityRef(self, context, base, sysid, pubid):
        self.origReader.entityRef(self.enc(context), self.enc(base), self.enc(sysid), self.enc(pubid))
        return

    def resolveEntity(self, pubid, sysid):
        self.origReader.resolveEntity(self.enc(sysid), self.enc(pubid))
        return


### Domlette Parser Interface ###

class PyExpatReader(ReaderBase.DomletteReader):
    def __init__(self, override=0, resolveEntity=None, force8Bit=0):
        '''
        override means that this class is not really doing the parsing, but
        merely providing the handlers and state for the parsing code
        '''
        ReaderBase.DomletteReader.__init__(self, force8Bit)
        self._override = override
        self._resolveEntity=None
        return

    def releaseNode(self, node):
        ReleaseNode(node)

    def getRootNode(self):
        return self._rootNode or self._ownerDoc
    
    def fromStream(self, stream, baseUri='',ownerDoc=None, stripElements=None):
        #Override fromStream as required
        if not self._override:
            self.initParser()
        self.initState(ownerDoc, stripElements)
        p = self.parser
        p.SetBase(baseUri)
        try:
            self.parser.ParseFile(stream)
        except:
            if self._rootNode: ReleaseNode(self._rootNode)
            if self._ownerDoc: ReleaseNode(self._ownerDoc)
            raise FtException(Error.XML_PARSE_ERROR, p.ErrorLineNumber, p.ErrorColumnNumber, expat.ErrorString(p.ErrorCode))
        if not self._override:
            self.killParser()
        self._completeTextNode()
        return self.getRootNode()

    ################
    #pDomlette specific interfaces
    ################

    def initState(self, ownerDoc=None, stripElements=None):
        self._preserveStateStack = [1]
        self._stripElements = stripElements or []
        if ownerDoc:
            self._ownerDoc = ownerDoc
            #Create a docfrag to hold all the generated nodes.
            self._rootNode = DocumentFragment(ownerDoc)
        else:
            self._rootNode = self._ownerDoc = Document()
        #Set up the stack which keeps track of the nesting of DOM nodes.
        self._nodeStack = [self._rootNode]
        self._namespaces = [{'xml': XML_NAMESPACE}]
        self._currText = ''
        return

    def initParser(self):
        if self.force8Bit:
            handler = Utf8OnlyHandler(self)
        else:
            handler = self
        self.parser=expat.ParserCreate()
        self.parser.StartElementHandler = handler.startElement
        self.parser.EndElementHandler = handler.endElement
        self.parser.CharacterDataHandler = handler.characters
        self.parser.ProcessingInstructionHandler = handler.processingInstruction
        self.parser.CommentHandler = handler.comment
        self.parser.ExternalEntityRefHandler = handler.entityRef
        return

    def killParser(self):
        self.parser = None

    def _completeTextNode(self):
        #Note some parsers don't report ignorable white space properly
        if self._currText and len(self._nodeStack) and self._nodeStack[-1].nodeType != Node.DOCUMENT_NODE:
            if self._preserveStateStack[-1] or string.strip(self._currText):
                new_text = Text(self._ownerDoc)
                new_text.data = self._currText
                self._nodeStack[-1].appendChild(new_text)
        self._currText = ''
        return

    def processingInstruction(self, target, data):
        self._completeTextNode()
        pi = ProcessingInstruction(self._ownerDoc)
        pi.target = target
        pi.data = data
        self._nodeStack[-1].appendChild(pi)

    def comment(self, data):
        self._completeTextNode()
        comment = Comment(self._ownerDoc)
        comment.data = data
        self._nodeStack[-1].appendChild(comment)

    def startElement(self, name, attribs):
        self._completeTextNode()
        self._namespaces.append(self._namespaces[-1].copy())
        for curr_attrib_key in attribs.keys():
            (prefix, local) = SplitQName(curr_attrib_key)
            if local == 'xmlns':
                self._namespaces[-1][prefix] = attribs[curr_attrib_key]
        (prefix, local) = SplitQName(name)

        if prefix and not self._namespaces[-1].has_key(prefix):
            raise "NAMESPACE EXCEPTION %s" % prefix

        new_element = Element(
            self._ownerDoc,
            self._namespaces[-1].has_key(prefix) and self._namespaces[-1][prefix] or '',
            local,
            prefix
            )

        for curr_attrib_key in attribs.keys():
            (prefix, local) = SplitQName(curr_attrib_key)
            if local == 'xmlns':
                node = Attribute(self._ownerDoc, XMLNS_NAMESPACE, prefix,
                                 local)
                new_element.attributes[(XMLNS_NAMESPACE,prefix)] = node
            else:
                if prefix and self._namespaces[-1].has_key(prefix):
                    ns = self._namespaces[-1][prefix]
                else:
                    ns = ''
                node = Attribute(self._ownerDoc, ns, local, prefix)
                new_element.attributes[(ns, local)] = node
            node.nodeValue = node.value = attribs[curr_attrib_key]
            node.ownerElement = new_element
        new_pstate = self._preserveStateStack[-1]
        for (uri, local, strip) in self._stripElements:
            if (uri, local) in [(new_element.namespaceURI, new_element.localName), ('', '*'), (new_element.namespaceURI, '*')]:
                new_pstate = not strip
                break
        self._preserveStateStack.append(new_pstate)
        self._nodeStack.append(new_element)
        return

    def endElement(self, name):
        self._completeTextNode()
        new_element = self._nodeStack[-1]
        del self._preserveStateStack[-1]
        del self._nodeStack[-1]
        del self._namespaces[-1]
        self._nodeStack[-1].appendChild(new_element)

    def characters(self, data):
        self._currText = self._currText + data
        return

    _entity_stack= None
    def entityRef(self,context,base,sysid,pubid):
        stream= self.resolveEntity(pubid,_combineSysid(base,sysid))

        if self._entity_stack is None: self._entity_stack= []
        self._entity_stack.append(self.parser)

        self.parser= self.parser.ExternalEntityParserCreate(context)
        self.parser.SetBase(_getBase(stream))
        self.parser.ParseFile(stream)

        self.parser= self._entity_stack.pop()
        stream.close()
        return 1

    def resolveEntity(self,pubid,sysid):
        if self._resolveEntity is None: source= sysid
        else: source= self._resolveEntity(pubid, sysid)

        if type(source) in stringTypes:
            source= urllib.urlopen(source)

        return source


#FIXME: build support for unparsed-entity-uri
try:
    from xml.sax import saxlib, saxexts
    from xml.dom.ext.reader import Sax2Lib
    class DomletteGenerator(saxlib.HandlerBase, Sax2Lib.LexicalHandler,
                            Sax2Lib.DTDDeclHandler, Sax2Lib.NamespaceHandler):
        def __init__(self, keepAllWs=0):
            self.evtHandler = PyExpatReader(None, 1)
            self._keepAllWs = keepAllWs
            return

        def initState(self, ownerDoc=None, stripElements=None):
            self.evtHandler.initState(ownerDoc, stripElements)
            return

        def getRootNode(self):
            self.evtHandler._completeTextNode()
            return self.evtHandler.getRootNode()            

        #Overridden DocumentHandler methods

        def startElement(self, name, attribs):
            self.evtHandler.startElement(name, attribs)
            return
        
        def endElement(self, name):
            self.evtHandler.endElement(name)
            return
        
        def processingInstruction(self, target, data):
            self.evtHandler.processingInstruction(target, data)
            return
        
        def characters(self, ch, start, length):
            self.evtHandler.characters(ch[start:start+length])
        
        def ignorableWhitespace(self, ch, start, length):
            """
            If 'keepAllWs' permits, add ignorable white-space as a text node.
            A Document node cannot contain text nodes directly.
            If the white-space occurs outside the root element, there is no
            place for it in the DOM and it must be discarded.
            """
            if self._keepAllWs:
                self.evtHandler.characters(ch, start, length)
            return

        #Overridden ErrorHandler methods
        #FIXME: How do we handle warnings?

        def error(self, exception):
            raise exception

        def fatalError(self, exception):
            raise exception

        #Overridden DTDHandler methods
        def notationDecl(self, name, publicId, systemId):
            pass

        def unparsedEntityDecl(self, name, publicId, systemId, notationName):
            if not hasattr(self.evtHandler._ownerDoc._unparsedEntities):
                self.evtHandler._ownerDoc._unparsedEntities = {}
            self.evtHandler._ownerDoc._unparsedEntities[name] = systemId
            return


    class SaxReader(ReaderBase.DomletteReader):
        def __init__(self, validate=0, keepAllWs=0, catName=None,
                     saxHandlerClass=DomletteGenerator, parser=None,
                     force8Bit=0):
            ReaderBase.DomletteReader.__init__(self, force8Bit)
            self._origParser = parser
            self._validate = validate
            self._keepAllWs = keepAllWs
            self._catName = catName
            self._saxHandlerClass = saxHandlerClass

        def __getinitargs__(self):
            return (self.validate,self._keepAllWs,self._catName,self._saxHandlerClass,self._parser)

        def releaseNode(self, node):
            ReleaseNode(node)

        def fromStream(self, stream, baseUri='', ownerDoc=None,
                       stripElements=None):
            self.initParser()
            self.handler._stripElements = stripElements
            self.handler.initState(ownerDoc=ownerDoc,
                                   stripElements=stripElements)
            self.parser.parseFile(stream)
            self.killParser()
            return self.handler.getRootNode()

        def initParser(self):
            self.parser = self._origParser or (self._validate and saxexts.XMLValParserFactory.make_parser()) or  saxexts.XMLParserFactory.make_parser()
            if self._catName:
                #set up the catalog, if there is one
                from xml.parsers.xmlproc import catalog
                cat_handler = catalog.SAX_catalog(self._catName, catalog.CatParserFactory())
                self.parser.setEntityResolver(self._cat_handler)
            self.handler = self._saxHandlerClass(self._keepAllWs)
            if self.force8Bit:
                self.handler = Utf8OnlyHandler(self.handler)
            self.parser.setDocumentHandler(self.handler)
            self.parser.setDTDHandler(self.handler)
            self.parser.setErrorHandler(self.handler)
            return

        def killParser(self):
            self.parser = None


except ImportError:
    pass


def PickleDocument(doc):
    stream = cStringIO.StringIO()
    pickler = cPickle.Pickler(stream, sys.platform[:3] != 'win')
    _PickleChildren(pickler, doc.childNodes)
    return stream.getvalue()

def PickleNode(node):
    stream = cStringIO.StringIO()
    pickler = cPickle.Pickler(stream, 1)
    pickler.dump(node)
    _PickleChildren(pickler, node.childNodes)
    return stream.getvalue()


# Helper function for pickling
def _PickleChildren(pickler, children):
    pickler.dump(len(children))
    for child in children:
        pickler.dump(child)
        _PickleChildren(pickler, child.childNodes)

import threading
g_lock = threading.Lock()
def UnpickleDocument(pickledXml):
    g_lock.acquire()
    try:
        doc = Document()
        stream = cStringIO.StringIO(pickledXml)
        unpickler = cPickle.Unpickler(stream)
        _UnpickleChildren(unpickler, doc)
        return doc
    finally:
        g_lock.release()

def UnpickleNode(pickledXml, doc=None):
    g_lock.acquire()
    try:
        doc = doc or Document()
        stream = cStringIO.StringIO(pickledXml)
        unpickler = cPickle.Unpickler(stream)
        topLevelNode = unpickler.load()
        doc.appendChild(topLevelNode)
        if topLevelNode.attributes:
            for attr in topLevelNode.attributes.values():
                attr.ownerDocument = topLevelNode.ownerDocument
        _UnpickleChildren(unpickler, topLevelNode)
        return topLevelNode
    finally:
        g_lock.release()

## Helper function for unpickling ##


def _UnpickleChildren(unpickler, node):
    children = unpickler.load()
    while children:
        child = unpickler.load()
        node.appendChild(child)
        if child.nodeType == Node.ELEMENT_NODE:
            for attr in child.attributes:
                attr.ownerDocument = child.ownerDocument
        _UnpickleChildren(unpickler, child)
        children = children - 1


def _combineSysid(base,sysid):
    return urllib.basejoin(base, sysid)


def _getBase(stream):
    #FIXME: This is an odd request.  What does it mean to get the "base"
    #Of an arbitrary URI?
    name = getattr(stream, 'url', None) or getattr(stream, 'name', '')
    return name

