############################################################################
#
# File Name: 		CollectionOfPersistentObjects.py
#
# Documentation:	http://docs.ftsuite.com/4ODS/CollectionOfPersistentObjects.py.html
#
"""
Base class for all lists of persistent objects
WWW: http://4suite.com/4ODS         e-mail: support@4suite.com

Copyright (c) 1999 FourThought LLC, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""


from Ft.Ods import Database
from Ft.Ods import PersistentObject
from Ft.Ods.Collections import CollectionBase
from Ft.Ods.StorageManager.Adapters import Constants

Actions = Constants.CollectionChanges

import types


g_modifyThreshold = 0.50  # percentage of changes allowed to original data

class CollectionOfPersistentObjects(CollectionBase.CollectionBase):

    def __init__(self,cid=None,subType=None, values=None):

        self.__dict__['_dbChanges'] = []
        if cid is not None:
            # Recreate an existing object
            # map values to a list of tuples:
            #  (db_idx, id, data)

            values = map(lambda idx,v:[idx,v,None], range(len(values)), values)
            self.__dict__['_dbSize'] = len(values)
        else:
            self.__dict__['_dbSize'] = 0

        CollectionBase.CollectionBase.__init__(self,cid,subType,values)

        #Always
        self.__dict__['_subType'] = Constants.Types.OBJECT

        return

    def __getattr__(self,name):
        #Support for the value attribute
        if name == 'value':
            return map(lambda index,self=self:self._4ods_verifyLoaded(index),range(len(self._values)))
        raise AttributeError(name)

    #Internal
    def _4ods_verifyLoaded(self,index):
        if self._values[index][2]:
            return self._values[index][2]
        elif self._values[index][1]:
            o = self._db._4ods_getPersistentObjectById(self._pType,self._values[index][1])
            self._values[index][2] = o
            return o
        else:
            print self._values[index]
            raise "Confused"

    #4ODS interfaces

    def _4ods_setTransaction(self, tx, unregister=0):
        CollectionBase._4ods_setTransaction(self, tx, unregister)
        if tx:
            self.__dict__['_dbChanges'] = []
        return


    def _4ods_getChanges(self):
        #Map the changes
        newChanges = []
        for mode,data in self._dbChanges:
            if mode in [Actions.APPEND,Actions.INSERT]:
                newEntry = (mode,[])
                for obj,index in data:
                    newEntry[1].append((obj._4ods_getId(),index))
                newChanges.append(newEntry)
            else:
                newChanges.append((mode,data))
        self.__dict__['_dbSize'] = len(self._values)
        self.__dict__['_dbChanges'] = []
        
        values = []
        for idx,oid,v in self._values:
            if oid:
                values.append((idx,oid,v))
            else:
                values.append((idx,v._4ods_getId(),v))
        self.__dict__['_values'] = values
        return newChanges
        #NOTE Optimize by checking size to num of operations
        #if not self._initialized:
        #    # Return the whole thing, not in the database yet
        #    changes = (None, contents)
        #elif len(self._dbChanges) < (self._dbSize * g_modifyThreshold):
        #    # Number of changes less than the limit, return just the changes
        #    changes = (self._dbChanges, None)
        #else:
        #    # Threshold reached, return all values
        #    changes = (None, contents)
        #self.__dict__['_initialized'] = 1
        #return changes

    def _4ods_isModified(self):
        return len(self._dbChanges) > 0

    def _4ods_getSubType(self):
        return self._subType



    #Override some of the pythonic interfaces
    def append(self, element):
        if not isinstance(element,PersistentObject.PersistentObject):
            raise TypeError('Expected Object, got %s' % str(type(element)))

        self._values.append([None, element._4ods_getId(),element])

        if len(self._dbChanges) and self._dbChanges[-1][0] == Actions.APPEND:
            self._dbChanges[-1][1].append((element,len(self._values)-1))
        else:
            self._dbChanges.append((Actions.APPEND, [(element,len(self._values)-1)]))
        return 

    def insert(self, i, element):
        if not isinstance(element,PersistentObject.PersistentObject):
            raise TypeError('Expected Object, got %s' % str(type(element)))

        self._values.insert(i, [None, element._4ods_getId(),element])

        if len(self._dbChanges) and self._dbChanges[-1][0] == Actions.INSERT:
            self._dbChanges[-1][1].append((element, i))
        else:
            self._dbChanges.append((Actions.INSERT, [(element, i)]))
        return 

    def remove(self, item):
        if not isinstance(item,PersistentObject.PersistentObject):
            raise TypeError('Expected Object, got %s' % str(type(element)))
        otherPid = item._4ods_getId()
        ctr = 0
        for (idx,pid,element) in self._values:
            if (otherPid and pid and otherPid == pid) or item == self._4ods_verifyLoaded(ctr):
                del self[ctr]
                #NOTE delitem adjusts the dbChanges
                break
            ctr = ctr + 1
        else:
            raise ValueError('%s not in collection' % str(element))

    def count(self, item):
        itemPid = item._4ods_getId()
        if itemPid:
            return len(filter(lambda (idx,pid,object),i=itemPid:pid==i,self._values))
        else:
            new = filter(lambda (idx,pid,object):not pid,self._values)
            return len(filter(lambda (idx, pid, object),i=item:object == i,new))

    def index(self, item):
        contents = map(lambda index,self=self:self._4ods_verifyLoaded(index),range(len(self._values)))
        return contents.index(item)

    def __getitem__(self, i):
        # Return the data portion of the tuple
        return self._4ods_verifyLoaded(i)

    def __setitem__(self, i, item):
        del self[i]
        self.insert(i,item)

    def __delitem__(self, i):
        value = self._values.pop(i)
        if len(self._dbChanges) and self._dbChanges[-1][0] == Actions.REMOVE:
            self._dbChanges[-1][1].append(i)
        else:
            self._dbChanges.append((Actions.REMOVE, [i]))


    def _pseudo_del(self):
        self.__dict__['_dbChanges'] = []
        CollectionBase.CollectionBase._pseudo_del(self)
