########################################################################
#
# File Name: 		Database.py
#
# Documentation:	http://docs.ftsuite.com/4ODS/Database.py.html
#
"""
Implements the Database interface.
WWW: http://4suite.com/4ODS         e-mail: support@4suite.com

Copyright (c) 1999-2000 FourThought Inc, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""

import string, threading
from Ft.Ods import Transaction
from Ft.Ods.Transaction import TransactionInProgress
from Ft.Ods.Transaction import TransactionNotInProgress
import Ft.Ods
from Ft.Ods.StorageManager.Adapters import Constants


# Exception classes
class DatabaseOpen:
    pass
    
class DatabaseClosed:
    pass
    
class DatabaseNotFound:
    pass
    
class ObjectNameNotUnique:
    pass
    
class ObjectNameNotFound:
    pass
    
def GetThreadId():
    return id(threading.currentThread())
#GetThreadId = thread.get_ident

g_db = None

def Database():
    global g_db
    if not g_db:
        g_db = _Database()
    return g_db
        
class _Database:
    
    def __init__(self):

        #A dictionary of current transactions
        #{TxId:transaction instance}
        self.__transactions = {}
        #A list of threads and transactions
        #{threadId:TxId}
        self.__threads = {}

        #Our thread safe lock
        #We must make sure that only one call at a time goes to our storage manager
        self.__lock = threading.RLock()

        #The name of the current database to be opened
        #{threadId:name}
        self.__dbName = None


    def new(self):
        tx = None
        try:
            self.__lock.acquire()

            #First up, do we know about this thread.
            #One transaction per thread only
            newThreadId = GetThreadId()
            if self.__threads.has_key(newThreadId):
                raise TransactionInProgress()
            if not self.__dbName:
                raise DatabaseClosed()
            tx = Transaction.Transaction(self,self.__dbName)
            txId = tx._4ods_getId()
            self.__transactions[txId] = tx
            self.__threads[newThreadId] = txId
        finally:
            self.__lock.release()
        return tx

        
    def open(self, database_name):
        try:
            self.__lock.acquire()
            if self.__dbName:
                raise DatabaseOpen()
            self.__dbName = database_name
        finally:
            self.__lock.release()
            
    def close(self):
        try:
            self.__lock.acquire()
            if not self.__dbName:
                raise DatabaseClosed()
            txIds = self.__transactions.keys()[:]

            #Abort any outstanding transactions
            for txId in txIds:
                self.__transactions[txId]._4ods_close()

            self.__dbName = None
            self.__transactions = {}
            self.__threads = {}

        finally:
            self.__lock.release()

    def bind(self, object, name):
        return self.__getCurrentTransaction()._4ods_bind(object,name)

    def unbind(self, name):
        return self.__getCurrentTransaction()._4ods_unbind(name)


    def lookup(self, name):
        return self.__getCurrentTransaction()._4ods_lookup(name)

    def keys(self,object=None):
        if object:
            return self.__getCurrentTransaction()._4ods_getObjectBindings(object)
        else:
            return self.__getCurrentTransaction()._4ods_getAllBindings()
        
    def isBound(self,object):
        return self.__getCurrentTransaction()._4ods_isBound(object)

    def schema(self):
        return self.__getCurrentTransaction()._4ods_getRepositoryObject(1)

    def _4ods_query(self,st):
        return self.__getCurrentTransaction()._4ods_runQuery(st)

    ## Private methods

    def __getCurrentTransaction(self):
        self.__lock.acquire()
        try:
            threadId = GetThreadId()
            if not self.__threads.has_key(threadId):
                raise TransactionNotInProgress()
            return self.__transactions[self.__threads[threadId]]
        finally:
            self.__lock.release()

    current = __getCurrentTransaction
            
    def __getCurrentTransactionId(self):
        self.__lock.acquire()
        try:
            threadId = GetThreadId()
            if not self.__threads.has_key(threadId):
                raise TransactionNotInProgress()
            return self.__threads[threadId]
        finally:
            self.__lock.release()

    def _4ods_inTransaction(self):
        return self.__threads.has_key(GetThreadId())

    def _4ods_isOpen(self):
        return self.__dbName != None
        
    ### Internal Callback Methods ###

    def _4ods_endTransaction(self, txId):
        self.__lock.acquire()
        try:
            del self.__transactions[txId]
            for tId in self.__threads.keys():
                if self.__threads[tId] == txId:
                    del self.__threads[tId]
            
        finally:
            self.__lock.release()

    def _4ods_getPersistentObjectById(self,_typ,id):
        tx = self.__getCurrentTransaction()
        if _typ == Constants.ObjectTypes.REPO:
            return tx._4ods_getRepositoryObject(id)
        elif _typ == Constants.ObjectTypes.OBJECT:
            return tx._4ods_getObject(id)
        elif _typ == Constants.ObjectTypes.COLLECTION:
            return tx._4ods_getCollection(id)
        else:
            raise "Unknown Type %s" % _typ

    def _4ods_loadExtent(self,id):
        tx = self.__getCurrentTransaction()
        return tx._4ods_loadExtent(id)

    def _4ods_loadCollectionData(self,c,index=None):
        tx = self.__getCurrentTransaction()
        return tx._4ods_loadCollectionData(c,index)

    def _4ods_registerThread(self,tx):
        #We want to register the current thread with the txId specified
        self.__lock.acquire()
        try:
            txId = tx._4ods_getId()
            threadId = GetThreadId()
            if self.__threads.has_key(threadId):
                raise TransactionInProgress()
            self.__threads[threadId] = txId
        finally:
            self.__lock.release()
            
    def _4ods_unregisterThread(self,tx):
        #We need to remove this transaction from the threads
        self.__lock.acquire()
        try:
            txId = tx._4ods_getId()
            threadId = GetThreadId()
            if not self.__threads.has_key(threadId):
                raise TransactionNotInProgress()
            del self.__threads[threadId]
        finally:
            self.__lock.release()
