############################################################################
#
# File Name: 		__init__.py
#
# Documentation:	http://docs.ftsuite.com/4ODS/StorageManager/__init__.py/html
#
"""
Manages data storage and transactions
WWW: http://4suite.com/4ODS         e-mail: support@4suite.com

Copyright (c) 2000 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""

import threading,types

from Ft.Ods.StorageManager import Adapters
from Ft.Ods.StorageManager import GlobalCache
from Ft.Ods.StorageManager.Adapters import ClassCache
from Ft.Ods.StorageManager import Extent
from Ft.Ods.StorageManager.Adapters import Constants
from Ft.Ods import FtodsBindingException
from Ft.Ods import FtodsObjectNotFound
import Ft.Ods
class TransactionInProgress:
    pass

class TransactionNotInProgress:
    pass


LEGAL_NAME_CHARS = "01234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_- /"

class StorageManager:
    TEST_MODE=0
    def __init__(self,factory,dbName):

        self._id = id(self)
        self._factory = factory
        self._open = 0
        self._hasBeenClosed = 0
        self._dbName = dbName
        

        #Create our adapter
        self._adapter = Adapters.GetAdapter()
        self._manager = Adapters.GetManager()
        self._gc = GlobalCache.GlobalCache()
        self._db = None

    #The transaction interface
    def begin(self):
        if self._open or self._hasBeenClosed:
            raise TransactionInProgress()
        self._open = 1
        self._4ods_clear()
        self._db = self._manager.connect(self._dbName)
        self._adapter.begin(self._db)

    def abort(self):
        if not self._open:
            raise TransactionNotInProgress()
        self._factory._4ods_endTransaction(self._id)
        self._adapter.abort(self._db)
        self._db = None
        self._open = 0
        self._hasBeenClosed = 1
        self._4ods_clean()

    def commit(self):
        if not self._open:
            raise TransactionNotInProgress()
        self._4ods_flush()
        self._open = 0
        self._hasBeenClosed = 1
        self._adapter.commit(self._db)
        self._db = None
        self._factory._4ods_endTransaction(self._id)
        self._4ods_clean()

    def checkpoint(self):
        if not self._open:
            raise TransactionNotInProgress()
        self._4ods_flush()
        self._adapter.checkpoint(self._db)


        #Change out the TX on the deleted objects
        for objs in self._deadObjects.values():
            for o in objs.values():
                o._4ods_setTransaction(None)


        #Clean up the deleted
        self._deadObjects = {}
        self._deadObjects[Constants.ObjectTypes.REPO] = {}
        self._deadObjects[Constants.ObjectTypes.OBJECT] = {}
        self._deadObjects[Constants.ObjectTypes.LITERAL] = {}
        self._deadObjects[Constants.ObjectTypes.COLLECTION] = {}

        #Clean up the new
        self._newObjects = {}
        self._newObjects[Constants.ObjectTypes.REPO] = []
        self._newObjects[Constants.ObjectTypes.OBJECT] = []
        self._newObjects[Constants.ObjectTypes.LITERAL] = []
        self._newObjects[Constants.ObjectTypes.COLLECTION] = []
            

    def join(self):
        if self._open:
            self._factory._4ods_registerThread(self)
        else:
            raise TransactionNotInProgress()

    def leave(self):
        if self._open:
            self._factory._4ods_unregisterThread(self)
        else:
            raise TransactionNotInProgress()


    ##################################
    #4ODS only interfaces
    ##################################


    #Called by the database to any outstanding transactions
    #When the database gets closed
    def _4ods_close(self):
        self._adapter.abort(self._db)
        self._db = None
        self._open = 0
        self._hasBeenClosed = 1
        self._4ods_clean()

    #Called to send the currently modified objects to disk
    def _4ods_flush(self):
        if not self._open:
            raise TransactionNotInProgress()


        keys = self._newObjects.keys()
        #Sort so that repos are done first
        keys.sort()

        #Have all of the blobs updated
        for (owner,name),data in self._newBlobs.items():
            bid = self._adapter.newBlob(self._db)
            owner._4ods_setBlobId(name,bid)
            self._adapter.writeBlob(self._db,bid,data)
        for bid,data in self._modifiedBlobs.items():
            self._adapter.writeBlob(self._db,bid,data)
        for bid in self._deadBlobs.keys():
            self._adapter.deleteBlob(self._db,bid)



        #New IDS
        for typ in keys:
            for obj in self._newObjects[typ]:
                self.newIdMapping[typ](self,obj)


        #Needed by some adapters
        self._newObjects[Constants.ObjectTypes.REPO] = self._4ods_sortNewRepos(self._newObjects[Constants.ObjectTypes.REPO])

        #New Writes
        for typ in keys:
            for obj in self._newObjects[typ]:
                self.newWriteMapping[typ](self,obj)

        #Modified Writes
        for typ in keys:
            for id_,obj in self._modifiedObjects[typ].items():
                if obj._4ods_isModified():
                    self.writeMapping[typ](self,obj)

        #Dead Objects
        for typ in keys:
            for id_,obj in self._deadObjects[typ].items():
                self.deleteMapping[typ](self,obj)

        #Update bindings
        for name in self._deadBindings:
            self._adapter.unbind(self._db,name)
        for name in self._newBindings:
            self._adapter.bind(self._db,self._bindings[name]._4ods_getId(),name)

        #Have all extents update
        for name,ext in self._extents.items():
            ext._4ods_updateChanges(self._adapter,self._db)

    def _4ods_isOpen(self):
        return self._open


    def _4ods_getId(self):
        return self._id


    def _4ods_clear(self):

        self._newObjects = {}
        self._newObjects[Constants.ObjectTypes.REPO] = []
        self._newObjects[Constants.ObjectTypes.OBJECT] = []
        self._newObjects[Constants.ObjectTypes.LITERAL] = []
        self._newObjects[Constants.ObjectTypes.COLLECTION] = []

        self._modifiedObjects = {}
        self._modifiedObjects[Constants.ObjectTypes.REPO] = {}
        self._modifiedObjects[Constants.ObjectTypes.OBJECT] = {}
        self._modifiedObjects[Constants.ObjectTypes.LITERAL] = {}
        self._modifiedObjects[Constants.ObjectTypes.COLLECTION] = {}

        self._deadObjects = {}
        self._deadObjects[Constants.ObjectTypes.REPO] = {}
        self._deadObjects[Constants.ObjectTypes.OBJECT] = {}
        self._deadObjects[Constants.ObjectTypes.LITERAL] = {}
        self._deadObjects[Constants.ObjectTypes.COLLECTION] = {}

        self._extents = {}
        self._bindings = {}
        

        self._newBindings = []
        self._deadBindings = []


        self._newBlobs = {}
        self._deadBlobs = {}
        self._modifiedBlobs = {}


    def _4ods_clean(self):
        #Should we Free the repo objects? Yes
        for ro in self._newObjects[Constants.ObjectTypes.REPO] + self._modifiedObjects[Constants.ObjectTypes.REPO].values() + self._deadObjects[Constants.ObjectTypes.REPO].values():
            Ft.Ods.FreePersistentObject(ro)
        self._4ods_clear()


    def _4ods_registerNewPersistentObject(self,po):
        if not self._open:
            raise TransactionNotInProgress()
        self._newObjects[po._persistentType].append(po)

        #We need to update the extents of this object
        for name in po._extents:
            ext = self._4ods_lookup(name)
            ext._4ods_addObject(po)

    def _4ods_registerNewBlob(self,inst,name,data):
        if not self._open:
            raise TransactionNotInProgress()
        self._newBlobs[(inst,name)] = data
            

    def _4ods_registerModifiedPersistentObject(self,po):
        if not self._open:
            raise TransactionNotInProgress()
        self._modifiedObjects[po._persistentType][po._4ods_getId()] = po


    def _4ods_registerModifiedBlob(self,bid,data):
        if not self._open:
            raise TransactionNotInProgress()
        self._modifiedBlobs[bid] = data
        

    def _4ods_unregisterPersistentObject(self,po):
        if not self._open:
            raise TransactionNotInProgress()
        typ = po._persistentType
        id_ = po._4ods_getId()
        if self._modifiedObjects[typ].has_key(id_):
            del self._modifiedObjects[typ][id_]
        if self._deadObjects[typ].has_key(id_):
            del self._deadObjects[typ][id_]
        while po in self._newObjects[typ]:
            self._newObjects[typ].remove(po)

    def _4ods_deletePersistentObject(self, po):
        if not self._open:
            raise TransactionNotInProgress()

        if not po._4ods_getId():
            self._4ods_unregisterPersistentObject(po)
        else:
            self._deadObjects[po._persistentType][po._4ods_getId()] = po

        #We need to update the extents of this object
        for name in po._extents:
            ext = self._4ods_lookup(name)
            ext._4ods_removeObject(po)
        if po._persistentType == Constants.ObjectTypes.OBJECT:
            #Remove bindings
            for name in self._4ods_getObjectBindings(po):
                self._4ods_unbind(name)

    def _4ods_deleteBlob(self,inst,name,bid):
        if not self._open:
            raise TransactionNotInProgress()
        if not bid:
            if self._newBlobs.has_key((inst,name)):
                del self._newBlobs[(inst,name)]
        else:
            if self._modifiedBlobs.has_key(bid):
                del self._modifiedBlobs[bid]
            self._deadBlobs[bid] = 1
        

    def _4ods_getRepositoryObject(self,rid):
        if not self._open:
            raise TransactionNotInProgress()
        if self._modifiedObjects[Constants.ObjectTypes.REPO].has_key(rid):
            return self._modifiedObjects[Constants.ObjectTypes.REPO][rid]
        data = self._gc.getObject(self._adapter,self._db,Constants.ObjectTypes.REPO,rid)
        if not data:
            data = self._adapter.getRepositoryObject(self._db,rid)
            if data is None:
                raise FtodsObjectNotFound(rid)
            self._gc.registerObject(Constants.ObjectTypes.REPO,rid,data)
        return data[0](data[1])

    def _4ods_getObject(self,oid):
        if not self._open:
            raise TransactionNotInProgress()
        if self._modifiedObjects[Constants.ObjectTypes.OBJECT].has_key(oid):
            return self._modifiedObjects[Constants.ObjectTypes.OBJECT][oid]
        data = self._gc.getObject(self._adapter,self._db,Constants.ObjectTypes.OBJECT,oid)
        if not data:
            data = self._adapter.getObject(self._db,oid)
            if data is None:
                raise FtodsObjectNotFound(oid)
            self._gc.registerObject(Constants.ObjectTypes.OBJECT,oid,data)
        return data[0](data[1])

    def _4ods_getCollection(self,cid):
        if not self._open:
            raise TransactionNotInProgress()
        if self._modifiedObjects[Constants.ObjectTypes.COLLECTION].has_key(cid):
            return self._modifiedObjects[Constants.ObjectTypes.COLLECTION][cid]
        data = self._gc.getObject(self._adapter,self._db,Constants.ObjectTypes.COLLECTION,cid)
        if not data:
            data = self._adapter.getCollection(self._db,cid)
            if data is None:
                raise FtodsObjectNotFound(cid)
            self._gc.registerObject(Constants.ObjectTypes.COLLECTION,cid,data)
        return data[0](cid=cid,subType=data[1],values=data[2])

    def _4ods_readBlob(self,bid):
        rt = self._adapter.readBlob(self._db,bid)
        if rt == None:
            raise "Invalid Blob"
        return rt


    def _4ods_lookup(self,name):
        if not self._open:
            raise TransactionNotInProgress()
        if self._extents.has_key(name):
            return self._extents[name]
        elif self._bindings.has_key(name):
            return self._bindings[name]
        elif name not in self._deadBindings:
            #Get it from the adapter
            oid = self._adapter.lookup(self._db,name)
            if oid:
                obj = self._4ods_getObject(oid)
                self._bindings[name] = obj
                return obj
            ext = self._adapter.loadExtent(self._db,name)
            if ext is None:
                return None

            ext = Extent.Extent(name,self,ext)
            self._extents[name] = ext
            return ext


    def _4ods_bind(self,object,name):
        if not self._open:
            raise TransactionNotInProgress()
        #Make sure the name does not have any illegal characters
        if not type(name) == types.StringType:
            raise FtodsBindingException(name)
        if len(filter(lambda x, lc=LEGAL_NAME_CHARS:x not in lc,name)):
            raise FtodsBindingException(name)

        if self._4ods_lookup(name):
            raise FtodsBindingException(name)
        self._newBindings.append(name)
        self._bindings[name] = object


    def _4ods_unbind(self,name):
        if not self._open:
            raise TransactionNotInProgress()
        if self._bindings.has_key(name):
            del self._bindings[name]
        if name in self._newBindings:
            self._newBindings.remove(name)
        self._deadBindings.append(name)

    def _4ods_getAllBindings(self):
        if not self._open:
            raise TransactionNotInProgress()
        #Always go to the database (since we gotta anyways)
        bindings = self._adapter.getAllBindings(self._db) + self._newBindings
        for db in self._deadBindings:
            while db in bindings:
                bindings.remove(db)

        eNames = self._adapter.extentNames(self._db)

        newExts = filter(lambda x:x._new == 1,self._extents.values())
        deadExts = filter(lambda x:x._delete == 1,self._extents.values())

        eNames = eNames + newExts
        for de in deadExts:
            while de in eNames:
                eNames.remove(de)

        return bindings + eNames

    def _4ods_getObjectBindings(self,object):
        if not self._open:
            raise TransactionNotInProgress()

        bindings = self._adapter.getObjectBindings(self._db,object._4ods_getId())
        for name in self._newBindings:
            obj = self._bindings[name]
            if obj == object:
                bindings.append(name)

        for db in self._deadBindings:
            while db in bindings:
                bindings.remove(db)
        return bindings
            
    def _4ods_isBound(self,object):
        if not self._open:
            raise TransactionNotInProgress()
        return self._adapter.lookup(self._db,object._4ods_getId()) != 0

    def _4ods_runQuery(self,cmd):
        raise "FIXME"
        



    #The callbacks that do stuff
    def _4ods_newRepoObjectId(self,ro):
        #Give IDS to all new repo objects
        rid = None
        if ro.meta_kind._v in [0,1,8,15]:
            rid = hash(ro.absolute_name())
            if rid < 0: rid = rid*-1
            if rid == 0: rid = None

        cid = self._adapter.getPythonClassId(self._db,ro.__class__)
        if cid is None:
            self._adapter.newPythonClass(self._db,ro.__class__)

        rid = self._adapter.newRepositoryObjectId(self._db,ro.meta_kind._v,ro.__class__,rid=rid)
        ro._4ods_setRepositoryId(rid)


    def _4ods_newCollectionId(self,col):
        if not col._4ods_isModified():
            return
        kid = self._adapter.getPythonClassId(self._db,col.__class__)
        if kid is None:
            self._adapter.newPythonClass(self._db,col.__class__)

        cid = self._adapter.newCollection(self._db,col.__class__,col._collectionType,col._4ods_getSubType())
        col._4ods_setCid(cid)


    def _4ods_newObjectId(self,o):
        #Give IDS to all new objects
        cid = self._adapter.getPythonClassId(self._db,o.__class__)
        if cid is None:
            self._adapter.newPythonClass(self._db,o.__class__)

        oid = self._adapter.newObjectId(self._db,o.__class__)
        o._4ods_setObjectId(oid)


    
    def _4ods_newRepoObjectWrite(self,ro):
        #Write a new repo objects

        #Create storage for all new repo objects
        if ro.meta_kind._v in [8,1]:
            #Create a new storage class
            self._adapter.newInterfaceStorage(self._db,ro)
        #if ro.meta_kind._v == 2 and ro.collection_kind._v == 4:
        #    self._adapter._4ods_createDictionaryStorage(ro)

        #See if it defines an extent
        if ro.meta_kind._v == 1:
            for extName in ro.extents:
                self._extents[extName] = Extent.Extent(extName,self)


        self._adapter.writeRepositoryObject(self._db,ro.meta_kind._v,ro._4ods_getId(),ro._tupleTypes,ro._tupleNames,ro._4ods_getFullTuple())

    def _4ods_newObjectWrite(self,o):
        self._adapter.writeObject(self._db,o._4ods_getId(),o._typeId,o._tupleTypes,o._tupleNames,o._tupleIds,o._4ods_getFullTuple())

    def _4ods_newCollectionWrite(self,col):
        if col._4ods_getId():
            self._adapter.writeCollection(self._db,col._4ods_getId(),col._4ods_getChanges(),col._collectionType,col._4ods_getSubType())

        
    def _4ods_writeRepoObject(self,ro):
        #Write a new repo objects
        self._adapter.writeRepositoryObject(self._db,ro.meta_kind._v,ro._4ods_getId(),ro._tupleTypes,ro._tupleNames,ro._4ods_getModifiedTuple())

    _4ods_writeCollection = _4ods_newCollectionWrite
    _4ods_writeObject = _4ods_newObjectWrite


    def _4ods_deleteRepoObject(self,ro):
        #Write a new repo objects
        self._adapter.deleteRepositoryObject(self._db,ro._4ods_getId(),ro.meta_kind._v)
        #See if it defines an extent
        if ro.meta_kind._v == 1:
            for extName in ro.extents:
                ext = self._4ods_lookup(extName)
                ext.delete()

    def _4ods_deleteObject(self,o):
        #Write a new repo objects
        self._adapter.deleteObject(self._db,o._4ods_getId(),o.__class__)


    def _4ods_deleteCollection(self,col):
        self._adapter.deleteCollection(self._db,col._4ods_getId(),col._collectionType,col._4ods_getSubType())

    newIdMapping = {}
    newIdMapping[Constants.ObjectTypes.REPO] = _4ods_newRepoObjectId
    newIdMapping[Constants.ObjectTypes.OBJECT] = _4ods_newObjectId
    newIdMapping[Constants.ObjectTypes.COLLECTION] = _4ods_newCollectionId



    newWriteMapping = {}
    newWriteMapping[Constants.ObjectTypes.REPO] = _4ods_newRepoObjectWrite
    newWriteMapping[Constants.ObjectTypes.OBJECT] = _4ods_newObjectWrite
    newWriteMapping[Constants.ObjectTypes.COLLECTION] = _4ods_newCollectionWrite

    writeMapping = {}
    writeMapping[Constants.ObjectTypes.REPO] = _4ods_writeRepoObject
    writeMapping[Constants.ObjectTypes.OBJECT] = _4ods_writeObject
    writeMapping[Constants.ObjectTypes.COLLECTION] = _4ods_writeCollection

    deleteMapping = {}
    deleteMapping[Constants.ObjectTypes.REPO] = _4ods_deleteRepoObject
    deleteMapping[Constants.ObjectTypes.OBJECT] = _4ods_deleteObject
    deleteMapping[Constants.ObjectTypes.COLLECTION] = _4ods_deleteCollection



    def _4ods_sortNewRepos(self,repos):

        res = []
        candi = {}
        for r in repos:
            if r.meta_kind._v == 8:
                candi[r] = []
                for i in r.inherits:
                    candi[r].append(i)
            elif r.meta_kind._v == 1:
                candi[r] = []
                for i in r.inherits:
                    candi[r].append(i)
                if r.extender:
                    candi[r].append(r.extender)
            else:
                res.append(r)

        tables = []
        for r,derived in candi.items():
            curIndex = 0
            for d in derived:
                if d in tables:
                    curIndex = tables.index(d)+1
            tables.insert(curIndex,r)
        return res + tables
