########################################################################
#
# File Name:            Model.py
#
# Documentation:        http://docs.4suite.com/4Rdf/Model.py.html
#
"""
The model container for RDF meta-data: represents a directed graph
WWW: http://4suite.com/4RDF         e-mail: support@4suite.com

Copyright (c) 1999 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""

from Ft.Rdf import Resource
from Ft.Rdf import Statement
from Ft.Lib import Uuid
from Ft.Rdf import Container
from Ft.Rdf import RDF_MS_BASE, RDF_SCHEMA_BASE
from types import *

class Model(Resource.Resource):
    def __init__(self, sourceUri, driver, schemaHandler=None):
        Resource.Resource.__init__(self, sourceUri)
        self.sourceUri = sourceUri
        self._driver = driver
        self._schemaHandler = schemaHandler
        if schemaHandler:
            schemaHandler.initializeModel(self)
        return

    def addContainer(self, container):
        uri = container.uri
        if not uri:
            #We need to generate one
            uri = self.generateUri()
            
        ctype = container.className
        stmts = [(uri, RDF_MS_BASE + 'type', RDF_MS_BASE+ctype, uri)]
        idx = 1
        for item in container:
            stmts.append((uri, RDF_MS_BASE + '_%d' % idx, item, uri))
            idx = idx + 1
        self._driver.add(stmts)
        return uri

    def extractContainer(self, uri):
        #Get all of the info about this ID
        rt = self.complete(uri, None, None)
        #There are 2 types of statements we care about, the type and the _*s
        test = RDF_MS_BASE+'type'
        type_list = filter(lambda x, test=test: x.predicate == test, rt)

        if not len(type_list):
            return None

        ctype = type_list[0].object[len(RDF_MS_BASE):]

        test = RDF_MS_BASE+'_'
        contents = filter(lambda x, test=test:
                          x.predicate[:len(test)] == test, rt)

        #Make sure there is only on copy of each _
        newContents = []
        found = []
        for c in contents:
            if c.predicate not in found:
                newContents.append(c)
                found.append(c.predicate)
        contents = newContents

        if ctype == 'Seq':
            # Remove the leading underscore also
            strip_len = len(RDF_MS_BASE) + 1
            contents.sort(lambda left, right, strip=strip_len:
                          cmp(int(left.predicate[strip:]),
                              int(right.predicate[strip:]))
                          )

        contents = map(lambda x:x.object,contents)

        cont = None
        if ctype == 'Bag':
            cont = Container.Bag(uri, contents)
        elif ctype == 'Seq':
            cont = Container.Sequence(uri, contents)
        elif ctype == 'Alt':
            cont = Container.Alternative(uri, contents)
        else:
            #Invalid??
            return None
        return cont



    def add(self, stmt, checkSchema=1):
        if type(stmt) not in [ListType, TupleType]:
            stmt = (stmt,)
        if self._schemaHandler and checkSchema:
            for s in stmt:
                self._schemaHandler.checkConstraints(self, s)
        self._driver.add(self._unmapStatements(stmt))

    def remove(self, stmt):
        if type(stmt) not in [ListType, TupleType]:
            stmt = (stmt,)
        self._driver.remove(self._unmapStatements(stmt))
        return

    def removeAll(self, uri=None):
        self._driver.removeAll(uri)
        return

    def contains(self, stmt):
        return apply(self._driver.contains,self._unmapStatements([stmt])[0])

    def statements(self):
        return self.complete(None, None, None)

    def size(self):
        return self._driver.size()

    def complete(self, subj, pred, obj):
        """
        Any combination of s, p and o can be None, and any None slot is
        treated as a wildcard that matches any value in the model.  A list
        is returned comprising all the statement objects that match the
        arguments.
        """
        rt = self._driver.complete(subj, pred, obj)
        return self._mapStatements(rt)

    def completeRegex(self, subj, pred, obj, ignoreCase=0):
        """
        One step up from the complete, allows you to use regular expressions
        in the subject, predicate, and object
        """
        rt = self._driver.completeRegex(subj, pred, obj,ignoreCase)
        return self._mapStatements(rt)

    def exclude(self, subj, pred, obj):
        rt = self._driver.exclude(subj, pred, obj)
        return self._mapStatements(rt)

    def generateUri(self):
        return 'urn:uuid:'+Uuid.UuidAsString(Uuid.GenerateUuid())

    def bind(self, object, name):
        self._driver.bind(object, name)

    def unbind(self, name):
        self._driver.unbind(name)

    def lookup(self, name):
        return self._driver.lookup(name)

    def has_key(self,name):
        return self._driver.has_key(name)

    def compileRil(self,exp):
        from Ft.Rdf.Parsers.Ril import CompiledRilExpression
        return CompiledRilExpression.CompiledRilExpression(exp)

    def _dump(self):
        print "Model: %s" % self.sourceUri
        for s in self.statements():
            print s

    def _mapStatements(self,statements):
        return map(lambda x: Statement.Statement(x[0],x[1],x[2],x[3]), statements)

    def _unmapStatements(self,statements):
        return map(lambda x: (x.subject,x.predicate,x.object,x.uri), statements)
        
