import sys, traceback, os, cStringIO, time, string
import linecache

OK = 0
PASSED = 1
FAILED = -1

class TestItem:
    def __init__(self, suite, title, group):
        self.suite = suite
        self.title = title
        self.group = group
        self.messages = []
        self.hasErrors = 0
        self.hasWarnings = 0
        self.startTime = time.time()
        self.startCompare = 0

        if suite.useColor:
            #self.moveTo = '\033[%dG' % (suite.columns - 9)
            self.colorSuccess = '\033[1;32m'
            self.colorFailure = '\033[1;31m'
            self.colorWarning = '\033[1;33m'
            self.colorNormal = '\033[0;39m'
        else:
            self.colorSuccess = ''
            self.colorFailure = ''
            self.colorWarning = ''
            self.colorNormal = ''

    def done(self):
        return self.group.testDone()

    def finishRun(self):
        self.startCompare = time.time()
        return

    def finish(self):
        if self.hasErrors:
            msg = self._failure()
            retVal = FAILED
        elif self.hasWarnings:
            msg = self._passed()
            retVal = PASSED
        else:
            msg = self._success()
            retVal = OK

        title = self.title + ' ('
        if self.startCompare:
            title = title + '%1.3lf/' % (self.startCompare - self.startTime)
        title = title + '%1.3lf secs)' % (time.time() - self.startTime)
        spaces = self.suite.columns - 9
        spaces = spaces - len(title)
        print title + ' '*spaces + msg

        for msg in self.messages:
            print msg
        self.suite = None
        self.group = None
        return retVal

    def debug(self, msg):
        self.messages.append(msg)

    def message(self, msg):
        self.messages.append(msg)

    def warning(self, msg=None):
        msg and self.messages.append(msg)
        self.hasWarnings = 1

    def error(self, msg=None, saveTrace=0):
        filename = None
        lineno = None
        msgs = []
        msg and msgs.append(msg)
        if saveTrace:
            stream = cStringIO.StringIO()
            traceback.print_exc(None, stream)
            msgs.append(stream.getvalue())

        if filename is None:
            try:
                raise "Foo"
            except:
                tb = sys.exc_info()[2]
                f = tb.tb_frame.f_back.f_back
                lineno = f.f_lineno
		co = f.f_code
		filename = co.co_filename                
                while string.find(filename,"TestSuite.") != -1:
                    f = f.f_back
                    lineno = f.f_lineno
                    co = f.f_code
                    filename = co.co_filename                
                    
		line = linecache.getline(filename, lineno)
                msgs.insert(0,"Line: %d File %s" % (lineno,filename))
                msgs.insert(1,"%s" % (line))

        for m in msgs:
            self.messages.append(m)

        self.hasErrors = 1
        if self.suite.stopOnError:
            retval = self.finish()
            sys.exit(retval)

    ### Internal Methods ###
    def _success(self):
        return '[%s  OK  %s]' % (self.colorSuccess, self.colorNormal)

    def _passed(self):
        return '[%sPASSED%s]' % (self.colorWarning, self.colorNormal)

    def _failure(self):
        return '[%sFAILED%s]' % (self.colorFailure, self.colorNormal)


class TestGroup:
    def __init__(self, suite, title=None):
        self.suite = suite
        self.title = title
        self.tests = []
        self.retVal = OK

        if title:
            msg = '********** ' + str(title) + ' **********'
            print msg

    ### Methods ###

    def finish(self):
        for test in self.tests:
            self.retVal = self.retVal or test.finish()
        self.test = None
        self.suite = None
        return self.retVal
        
    def startTest(self, title):
        test = TestItem(self.suite, title, self)
        self.tests.append(test)
        return test

    def finishRun(self):
        if self.tests:
            self.tests[-1].finishRun()
        return

    def testDone(self):
        if self.tests:
            self.retVal = self.retVal | self.tests[-1].finish()
            del self.tests[-1]
            return self.retVal


class TestSuite:
    def __init__(self, stopOnError=1, useColor=0, cols=80):
        self.useColor = useColor and (os.name == 'posix'
                                      and sys.stdout.isatty())
        self.stopOnError = stopOnError
        self.columns = cols
        self.groups = []
        self.retVal = OK
        self._compareCtr = 0

    def __del__(self):
        retVal = OK
        while len(self.groups):
            retVal = retVal or group[-1].finish()
        return retVal

    ### Methods ###

    def startGroup(self, title):
        group = TestGroup(self, title)
        self.groups.append(group)

    def groupDone(self):
        retVal = OK
        if self.groups:
            retVal = self.groups[-1].finish()
            del self.groups[-1]
        return retVal

    def startTest(self, title):
        if not self.groups:
            self.startGroup(self)
            print 'Added (null) group'
        self.groups[-1].startTest(title)
        self._compareCtr = 0

    def testDone(self):
        if self.groups:
            self.groups[-1].testDone()
        self._compareCtr = 0

    def testResults(self, expected, actual, done=1, msg=None, func=None,diff=0):
        self.groups[-1].finishRun()
        func = func or cmp
        self._compareCtr = self._compareCtr + 1
        if msg is None:
            msg = "Test %d in of %s" % (self._compareCtr,self.groups[-1].tests[-1].title)
        if func(expected, actual):
            msg and self.message(msg)
            if diff:
                self.diff(expected,actual)
            self.error("Expected: %s\nActual: %s" % (
                str(expected),
                str(actual),
                ))
            
            return 0
        elif done:
            self.testDone()
        return 1

    def testException(self,func,args,eClass,eCodes=None):
        eCodes = eCodes or {}
        try:
            apply(func,args)
        except eClass,e:
            for attrName,value in eCodes.items():
                if hasattr(e,attrName) and getattr(e,attrName) == value:
                    continue
                self.error("Wrong Exception Attr %s == %s"%(attrName,str(value)),1)
        except:
            self.error("Wrong Exception Raised",1)
        else:
            self.error("No Exception Raised")

        
    def compare(self, expected, actual, msg=None, func=None):
        """
        func is a function that returns the same as builtin function cmp
        """
        return self.testResults(expected, actual, 0, msg, func)

    def compareIn(self,expected,actual,msg=None):
        """Test that actual is in expected"""
        def func(exp,act):
            return cmp(1,act in exp)
        return self.testResults(expected, actual, 0, msg, func)
        
    
    def message(self, msg):
        if self.groups:
            if self.groups[-1].tests:
                self.groups[-1].tests[-1].message(msg)

    def warning(self, msg):
        if self.groups:
            if self.groups[-1].tests:
                self.groups[-1].tests[-1].warning(msg)

    def error(self, msg, saveTrace=0):
        if self.groups:
            if self.groups[-1].tests:
                self.groups[-1].tests[-1].error(msg, saveTrace)


    def diff(self,expected,actual):
        import tempfile

        if sys.platform[:3] != 'win':

            sys.stdout.write("Diff of expected and actual\n")
            fn1 = tempfile.mktemp()
            fn2 = tempfile.mktemp()
            f1 = open(fn1,'w')
            f2 = open(fn2,'w')
            f1.write(str(expected))
            f2.write(str(actual))
            f1.close()
            f2.close()
        
            os.system("diff -c %s %s" % (fn1,fn2))

            os.unlink(fn1)
            os.unlink(fn2)
        

